/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2014 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#pragma once

//#define HASH_EVAL
#define GRAIN_SIZE 4

struct sPawnHashEntry {
  U64 pawnKey;
  int mgPawns;
  int egPawns;
  int mgPassers;
  int egPassers;
};

struct sEvalHashEntry {
  U64 key;
  int score;
};

#define PAWN_HASH_SIZE  512 * 512
#define EVAL_HASH_SIZE  512 * 512

struct sEvaluator {
private:
  int attScore[2];         // king attack scores
  int mgMobility[2];       // midgame mobility scores
  int egMobility[2];       // endgame mobility scores
  int mgMisc[2];           // miscelanneous midgame scores
  int egMisc[2];           // miscelanneous endgame scores
  int kingTropismMg[2];
  int kingTropismEg[2];
  int attOnPieces[2];
  int pawnScoreMg[2];      // midgame pawn structure scores
  int pawnScoreEg[2];      // endgame pawn structure scores
  int passerScoreMg[2];    // midgame passed pawn scores
  int passerScoreEg[2];    // endgame passed pawn scores 
  U64 bbPawnTakes[2];      // squares controlled by pawns, used in mobility eval (pawn eval uses only occupancy masks)
  U64 bbPawnCanTake[2];    // squares that can be controlled by pawns as they advance, used in outpost eval
  U64 bbDiagChecks[2];
  U64 bbStraightChecks[2];
  U64 bbKnightChecks[2];
  U64 bbMinorCoorAttacks[2]; // bitboard to detect coordinated attacks on enemy king
  U64 bbRookCoorAttacks[2];
  U64 bbAllAttacks[2];     // squares attacked by a side
  int attCount[2];         // attack counter based on square control
  int checkCount[2];       // check threat counter
  int attNumber[2];        // no. of pieces participating in the attack
  int attWood[2];
  int mgFact,  egFact;     // material-driven scaling factors
  int mgScore, egScore;    // partial midgame and endgame scores (to be scaled)

  sPawnHashEntry PawnTT[PAWN_HASH_SIZE]; // pawn transposition table
#ifdef HASH_EVAL
  sEvalHashEntry EvalTT[EVAL_HASH_SIZE]; // eval transposition table
#endif
  
  int GetMaterialScore(sPos *p);
  void AddMobility(int pc, int side, int cnt);
  void AddMisc(int side, int mg, int eg);
  void AddKingAttack(int side, int pc, int cnt);
  void AddPawnProperty(int pawnProperty, int side, int sq);
  void AddPasserScore(int pawnProperty, int side, int sq);
  int CheckmateHelper(sPos *p);
  void InitStaticScore(void);            
  void InitDynamicScore(sPos *p);            
  void SetScaleFactor(sPos *p);
  int SetDegradationFactor(sPos *p, int stronger);
  int Interpolate(void);
  void SinglePawnScore(sPos *p, int side); // eval_pawns.c
  void EvalPawnCenter(sPos *p, int side);  // eval_pawns.c
  void EvalPawns(sPos *p);				    // eval_pawns.c
  void ScoreN(sPos *p, int side);
  void ScoreB(sPos *p, int side);
  void ScoreR(sPos *p, int side);
  void ScoreQ(sPos *p, int side);
  void ScoreP(sPos *p, int side);
  void ScorePatterns(sPos *p, int side);
  void ScoreKingShield(sPos *p, int side);
  void ScoreKingAttacks(sPos *p, int side);
  void ScoreRelationToPawns(sPos *p, int sd, int pc, int sq);
  void ScoreKingTropism(int pc, int sq, int sd, int ksq);
  void ScoreHanging(sPos *p, int side);
  int  EvalKingFile(sPos * p, int side, U64 bbFile);
  int  EvalFileShelter(U64 bbOwnPawns, int side);
  int  EvalFileStorm(U64 bbOppPawns, int side);
  int  EvalTrappedKnight(sPos *p, int sd);
  int  EvalTrappedBishop(sPos *p, int sd);
  int  EvalTrappedRook(sPos *p, int sd);
  int  PullToDraw(sPos *p, int score);
  int  FinalizeScore(sPos *p, int score);
public:
  int Normalize(int val, int limit);
  void ScaleValue(int * value, int factor);
  void DebugPst(sPos *p);
  int ReturnFast(sPos *p);
  int ReturnFull(sPos *p, int alpha, int beta);
};

extern struct sEvaluator Eval;

int NotOnBishColor(sPos * p, int bishSide, int sq);
int BishopsAreDifferent(sPos * p);

int PcMatMinor(sPos *p, int side);
int PcMatKnight(sPos *p, int side);
int PcMatBishop(sPos *p, int side);
int PcMatRook(sPos *p, int side);
int PcMatRookMinor(sPos *p, int side);
int PcMatQueen(sPos *p, int side);
int PcMatQueenMinor(sPos *p, int side);
int PcMatBB(sPos *p, int side);
int PcMatNN(sPos *p, int side);
int PcMatBN(sPos *p, int side);
int PcMatNone(sPos * p, int side);
