/*
  Rodent, a UCI chess playing engine derived from Sungorus 1.4
  Copyright (C) 2009-2011 Pablo Vazquez (Sungorus author)
  Copyright (C) 2011-2014 Pawel Koziol

  Rodent is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published 
  by the Free Software Foundation, either version 3 of the License, 
  or (at your option) any later version.

  Rodent is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty 
  of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "bitboard.h"
#include "data.h"
#include "rodent.h"

void sPos::DoMove(int move, UNDO *u)
{
  int sd   = side;         // moving side 
  int op   = Opp(sd);
  int fsq  = Fsq(move);    // start square
  int tsq  = Tsq(move);    // target square
  int ftp  = Tp(pc[fsq]);  // moving piece
  int ttp  = Tp(pc[tsq]);  // captured piece

  U64 bbMove = SqBb(fsq) | SqBb(tsq); // optimization from Stockfish

  // save data for undoing a move
  u->ttp = ttp;
  u->castleFlags = castleFlags;
  u->epSquare = epSquare;
  u->reversibleMoves = reversibleMoves;
  u->hashKey = hashKey;
  u->pawnKey = pawnKey;

  repetitionList[head++] = hashKey;

  // update reversible move counter (zeroing is done on captures and pawn moves)
  reversibleMoves++;

  hashKey     ^= zobCastle[castleFlags];
  castleFlags &= castleMask[fsq] & castleMask[tsq];
  hashKey     ^= zobCastle[castleFlags];

  // clear en passant square
  if (epSquare != NO_SQ) {
    hashKey ^= zobEp[File(epSquare)];
    epSquare = NO_SQ;
  }

  // move a piece from start square
  pc[fsq]      = NO_PC;
  pc[tsq]      = Pc(sd, ftp);
  hashKey     ^= zobPiece[Pc(sd, ftp)][fsq] ^ zobPiece[Pc(sd, ftp)][tsq];
  if (ftp == P) {
	  reversibleMoves = 0;
	  pawnKey ^= zobPiece[Pc(sd, ftp)][fsq] ^ zobPiece[Pc(sd, ftp)][tsq];
  } 
  bbCl[sd]   ^= bbMove;
  bbTp[ftp]   ^= bbMove;
  pstMg[sd] += Data.pstMg[sd][ftp][tsq] - Data.pstMg[sd][ftp][fsq];
  pstEg[sd] += Data.pstEg[sd][ftp][tsq] - Data.pstEg[sd][ftp][fsq];

  // on a king move update king location data
  if (ftp == K) kingSquare[sd] = tsq;
  
  // capture
  if (ttp != NO_TP) {
	reversibleMoves = 0;
    hashKey ^= zobPiece[Pc(op, ttp)][tsq];
	if (ttp == P) 
		pawnKey     ^= zobPiece[Pc(op, ttp)][tsq];
    bbCl[op]        ^= SqBb(tsq);
    bbTp[ttp]       ^= SqBb(tsq);
	pcCount[op][ttp]--;
    pieceMat[op] -= Data.matValue[ttp];
	phase        -= Data.phaseValue[ttp]; 
    pstMg[op]    -= Data.pstMg[op][ttp][tsq];
	pstEg[op]    -= Data.pstEg[op][ttp][tsq];
  }
  
  switch (MoveType(move)) {

  case NORMAL:
    break;

  case CASTLE:
    if (tsq > fsq) {
      fsq += 3;
      tsq -= 1;
    } else {
      fsq -= 4;
      tsq += 1;
    }
    pc[fsq]   = NO_PC;
    pc[tsq]   = Pc(sd, R);
    hashKey   ^= zobPiece[Pc(sd, R)][fsq] ^ zobPiece[Pc(sd, R)][tsq];
    bbCl[sd]  ^= SqBb(fsq) | SqBb(tsq);
    bbTp[R]   ^= SqBb(fsq) | SqBb(tsq);
    pstMg[sd] += Data.pstMg[sd][R][tsq] - Data.pstMg[sd][R][fsq];
	pstEg[sd] += Data.pstEg[sd][R][tsq] - Data.pstEg[sd][R][fsq];
    break;

  case EP_CAP:
    tsq ^= 8;
    pc[tsq]   = NO_PC;
    hashKey  ^= zobPiece[Pc(op, P)][tsq];
	pawnKey  ^= zobPiece[Pc(op, P)][tsq];
    bbCl[op] ^= SqBb(tsq);
    bbTp[P]  ^= SqBb(tsq);
	pcCount[op][P]--;
	phase     -= Data.phaseValue[P];
    pstMg[op] -= Data.pstMg[op][P][tsq];
	pstEg[op] -= Data.pstEg[op][P][tsq];
    break;

  case EP_SET:
    tsq ^= 8;
     if (bbPawnAttacks[sd][tsq] & (bbCl[op] & bbTp[P]) ) {
      epSquare = tsq;
      hashKey ^= zobEp[File(tsq)];
    }
    break;

  // promotion:    (1) add promoted piece and add values associated with it
  case N_PROM:  // (2) remove promoted pawn and substract values associated with it 
  case B_PROM: 
  case R_PROM: 
  case Q_PROM:
    ftp = PromType(move);
    pc[tsq]   = Pc(sd, ftp);
    hashKey  ^= zobPiece[Pc(sd, P)][tsq] ^ zobPiece[Pc(sd, ftp)][tsq];
	pawnKey  ^= zobPiece[Pc(sd, P)][tsq];
    bbTp[P]  ^= SqBb(tsq);
    bbTp[ftp]^= SqBb(tsq);
	pcCount[sd][ftp]++;
	pcCount[sd][P]--;
	pieceMat[sd] += Data.matValue[ftp];
	phase        += Data.phaseValue[ftp]     - Data.phaseValue[P];
    pstMg[sd]    += Data.pstMg[sd][ftp][tsq] - Data.pstMg[sd][P][tsq];
	pstEg[sd]    += Data.pstEg[sd][ftp][tsq] - Data.pstEg[sd][P][tsq];
    break;

  }
  side ^= 1;
  hashKey ^= SIDE_RANDOM;
}

void sPos::DoNull(UNDO *u)
{
  u->epSquare = epSquare;
  u->hashKey  = hashKey;
  u->pawnKey  = pawnKey;
  repetitionList[head++] = hashKey;
  reversibleMoves++;

  if (epSquare != NO_SQ) {
    hashKey ^= zobEp[File(epSquare)];
    epSquare = NO_SQ;
  }

  side ^= 1;
  hashKey ^= SIDE_RANDOM;
}
